'use strict';

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; }; //======================//
//     KEY HANDLERS     //
//======================//
// All functions dealing with keypresses (listened to on the keydown event)
// are here, with specific implementations for most types of key

var _constants = require('./constants');

var _helpers = require('./helpers');

var _helpers2 = _interopRequireDefault(_helpers);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

module.exports = {

  /**
   * NUMBER HANDLER
   * @param {currentState} Information about current finput state
   * @param {keyInfo} Information about the pressed key
   * @param {options} Configuration options for the input
   */
  onNumber: function onNumber(currentState, keyInfo, options) {
    // Remove characters in current selection
    var tempCurrent = _helpers2.default.editString(currentState.value, '', currentState.caretStart, currentState.caretEnd);
    var tempNew = _helpers2.default.editString(currentState.value, keyInfo.keyName, currentState.caretStart, currentState.caretEnd);

    var allowedNumber = !(currentState.value[0] === '-' && currentState.caretStart === 0 && currentState.caretEnd === 0) && _helpers2.default.allowedZero(tempCurrent, keyInfo.keyName, currentState.caretStart, options) && _helpers2.default.allowedDecimal(tempNew, options);

    var newState = _extends({}, currentState);
    if (allowedNumber) {
      newState.value = _helpers2.default.editString(currentState.value, keyInfo.keyName, currentState.caretStart, currentState.caretEnd);
      newState.caretStart += 1;
    } else {
      newState.valid = false;
    }

    return newState;
  },

  /**
   * MINUS HANDLER
   * @param {currentState} Information about current finput state
   * @param {keyInfo} Information about the pressed key
   * @param {options} Configuration options for the input
   */
  onMinus: function onMinus(currentState, keyInfo, options) {
    var minusAllowed = currentState.caretStart === 0 && (currentState.value[0] !== '-' || currentState.caretEnd > 0) && options.range !== _constants.RANGE.POSITIVE;

    var newState = _extends({}, currentState);
    if (minusAllowed) {
      newState.value = _helpers2.default.editString(currentState.value, '-', currentState.caretStart, currentState.caretEnd);
      newState.caretStart += 1;
    } else {
      newState.valid = false;
    }

    return newState;
  },

  /**
   * DECIMAL HANDLER
   * @param {currentState} Information about current finput state
   * @param {keyInfo} Information about the pressed key
   * @param {options} Configuration options for the input
   */
  onDecimal: function onDecimal(currentState, keyInfo, options) {
    var decimalIndex = currentState.value.indexOf(options.decimal);

    // If there is not already a decimal or the original would be replaced
    // Add the decimal
    var decimalAllowed = options.scale > 0 && (decimalIndex === -1 || decimalIndex >= currentState.caretStart && decimalIndex < currentState.caretEnd);

    var newState = _extends({}, currentState);
    if (decimalAllowed) {
      newState.value = _helpers2.default.editString(currentState.value, options.decimal, currentState.caretStart, currentState.caretEnd);
      newState.caretStart += 1;
    } else {
      newState.valid = false;
    }

    return newState;
  },

  /**
   * THOUSANDS HANDLER
   * @param {currentState} Information about current finput state
   */
  onThousands: function onThousands(currentState, keyInfo, options) {
    var newState = _extends({}, currentState);
    newState.valid = false;
    return newState;
  },

  /**
   * SHORTCUT HANDLER
   * @param {currentState} Information about current finput state
   * @param {keyInfo} Information about the pressed key
   * @param {options} Configuration options for the input
   */
  onShortcut: function onShortcut(currentState, keyInfo, options) {
    var multiplier = options.shortcuts[keyInfo.keyName] || 1;
    var adjustedVal = _helpers2.default.editString(currentState.value, '', currentState.caretStart, currentState.caretEnd);
    var rawValue = (_helpers2.default.formattedToRaw(adjustedVal, options) || 1) * multiplier;

    var newState = _extends({}, currentState);
    if (multiplier) {
      // If number contains 'e' then it is too large to display
      if (rawValue.toString().indexOf('e') === -1) {
        newState.value = String(rawValue);
      }
      newState.caretStart = newState.value.length + Math.log10(1000);
    }

    return newState;
  },

  /**
   * BACKSPACE HANDLER
   * @param {currentState} Information about current finput state
   * @param {keyInfo} Information about the pressed key
   * @param {options} Configuration options for the input
   */
  onBackspace: function onBackspace(currentState, keyInfo, options) {
    var thousands = options.thousands;
    var firstHalf = void 0,
        lastHalf = void 0;

    var newState = _extends({}, currentState);
    if (currentState.caretStart === currentState.caretEnd) {
      if (keyInfo.modifierKey) {
        // If CTRL key is held down - delete everything BEFORE caret
        firstHalf = '';
        lastHalf = currentState.value.slice(currentState.caretStart, currentState.value.length);
        newState.caretStart = 0;
      } else {
        // Assume as there is a comma then there must be a number before it
        var caretJump = 1;

        caretJump = currentState.caretStart - caretJump >= 0 ? caretJump : 0;
        firstHalf = currentState.value.slice(0, currentState.caretStart - caretJump);
        lastHalf = currentState.value.slice(currentState.caretStart, currentState.value.length);
        newState.caretStart += -caretJump;
      }
    } else {
      // Same code as onDelete handler for deleting a selection range
      firstHalf = currentState.value.slice(0, currentState.caretStart);
      lastHalf = currentState.value.slice(currentState.caretEnd, currentState.value.length);
    }

    newState.value = firstHalf + lastHalf;

    return newState;
  },

  /**
   * DELETE HANDLER
   * @param {currentState} Information about current finput state
   * @param {keyInfo} Information about the pressed key
   * @param {options} Configuration options for the input
   */
  onDelete: function onDelete(currentState, keyInfo, options) {
    var thousands = options.thousands;
    var firstHalf = void 0,
        lastHalf = void 0;

    var newState = _extends({}, currentState);
    if (currentState.caretStart === currentState.caretEnd) {
      var nextChar = currentState.value[currentState.caretStart];

      if (keyInfo.modifierKey) {
        // If CTRL key is held down - delete everything AFTER caret
        firstHalf = currentState.value.slice(0, currentState.caretStart);
        lastHalf = '';
      } else {
        // Assume as there is a comma then there must be a number after it
        var thousandsNext = nextChar === thousands;

        // If char to delete is thousands and number is not to be deleted - skip over it
        newState.caretStart += thousandsNext ? 1 : 0;

        var lastHalfStart = newState.caretStart + (thousandsNext ? 0 : 1);
        firstHalf = currentState.value.slice(0, newState.caretStart);
        lastHalf = currentState.value.slice(lastHalfStart, currentState.value.length);
      }
    } else {
      // Same code as onBackspace handler for deleting a selection range
      firstHalf = currentState.value.slice(0, currentState.caretStart);
      lastHalf = currentState.value.slice(currentState.caretEnd, currentState.value.length);
    }

    newState.value = firstHalf + lastHalf;

    return newState;
  },

  /**
   * UNDO HANDLER
   * @param {currentState} Information about current finput state
   * @param {keyInfo} Information about the pressed key
   * @param {options} Configuration options for the input
   * @param {history} The history manager
   */
  onUndo: function onUndo(currentState, keyInfo, options, history) {
    var newState = _extends({}, currentState);
    newState.value = history.undo();
    newState.caretStart = newState.value.length;

    return newState;
  },
  /**
   * REDO HANDLER
   * @param {currentState} Information about current finput state
   * @param {keyInfo} Information about the pressed key
   * @param {options} Configuration options for the input
   * @param {history} The history manager
   */
  onRedo: function onRedo(currentState, keyInfo, options, history) {
    var newState = _extends({}, currentState);
    newState.value = history.redo();
    newState.caretStart = newState.value.length;

    return newState;
  },

  /**
   * UNKNOWN HANDLER
   * @param {currentState} Information about current finput state
   * @param {keyInfo} Information about the pressed key
   */
  onUnknown: function onUnknown(currentState, keyInfo) {
    // all printable characters have a key with length of 1 
    // if a character has got this far it is an invalid character
    var isInvalid = keyInfo.keyName.length === 1 && !keyInfo.modifierKey;
    var newState = _extends({}, currentState);
    newState.valid = !isInvalid;

    return newState;
  }
};